# Copyright (c) 2010-2012 Outpost Embedded, LLC
# This file is part of the pydbusclone software package.
#
# pydbusclone is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 2 as published by the
# Free Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import Queue, logging
from threading import RLock

import dbus

from dbusclone.util import CallbackList
from dbusclone.mainloop import call_in_main_thread
from dbusclone.exceptions import TimeoutError
from dbusclone.errors import convert_dbus_exc


DEFAULT_TIMEOUT = 60


logger = logging.getLogger(__name__)


class Object(object):
    bus = None
    ready_callback = None
    default_error_handler = None

    def __init__(self, bus, ready_callback, default_error_handler):
        if not callable(ready_callback):
            raise TypeError(
              'ready_callback is not callable: %r' % ready_callback
            )

        self.bus = bus
        self.ready_callback = ready_callback
        self.default_error_handler = default_error_handler

        self.maybe_ready()

    def is_ready(self):
        return True

    def maybe_ready(self):
        if self.is_ready():
            if callable(self.ready_callback):
                self.ready_callback(self)
                del self.ready_callback


class DBusObject(Object):
    bus_name = None
    object_path = None

    signal_matches = None

    def __init__(self, bus_name, object_path, **kwargs):
        self.bus_name = bus_name
        self.object_path = object_path

        super(DBusObject, self).__init__(**kwargs)

        self.signal_matches = set([])

    def __del__(self):
        if self.signal_matches:
            logger.error(
              'cleaning up %u signal matches for %r',
              len(self.signal_matches),
              self,
            )
            self.disconnect_signals()

        super_obj = super(DBusObject, self)
        if hasattr(super_obj, '__del__'):
            super_obj.__del__()

    def __repr__(self):
        return '<DBusObject "%s" "%s">' % (self.bus_name, self.object_path)

    # FIXME: dbus_interface
    def get_proxy_object(self):
        return self.bus.get_object(
          bus_name = self.bus_name,
          object_path = self.object_path,
        )

    def exists(self):
        # We actually do a synchronous call here just to determine if the
        # object exists.  The idea is that if the object does not exist,
        # our caller would want to know about that right away.
        #
        # My original thought was that Ping would be a good method to call
        # for this purpose, but it seems that people often forget to
        # allow calls to Ping in the D-Bus configuration, whereas Introspect
        # is more-or-less universally allowed.  I guess nobody uses Ping.
        try:
            dbus.Interface(
              self.get_proxy_object(),
              'org.freedesktop.DBus.Introspectable',
            ).Introspect()
        except dbus.DBusException:
            return False
        return True

    def call(self, method_name, *args, **kwargs):
        kwargs.setdefault('error_handler', self.default_error_handler)
        if not callable(kwargs['error_handler']):
            raise TypeError('error_handler must be callable')
        obj = self.get_proxy_object()
        fn = getattr(obj, method_name)
        call_in_main_thread(fn, *args, **kwargs)

    def call_and_wait(self, *args, **kwargs):
        '''Only useful from non-main threads.'''
        # FIXME: Check that we are not running in the main thread.
        queue = Queue.Queue()
        kwargs['reply_handler'] = (lambda *a: queue.put(a))
        kwargs['error_handler'] = queue.put
        self.call(*args, **kwargs)
        result = queue.get()
        if isinstance(result, Exception):
            if isinstance(result, dbus.DBusException):
                convert_dbus_exc(result)
            raise result
        # Don't return a tuple for functions that return a single argument.
        if len(result) == 1:
            return result[0]
        return result

    def connect_to_signal(self, *args, **kwargs):
        obj = self.get_proxy_object()
        signal_match = obj.connect_to_signal(*args, **kwargs)
        self.signal_matches.add(signal_match)

    def disconnect_signals(self):
        while self.signal_matches:
            self.signal_matches.pop().remove()


###


class PropertiesObject(Object):
    parent = None
    changed_callbacks = None

    def __init__(self, parent, **kwargs):
        self.parent = parent

        super(PropertiesObject, self).__init__(**kwargs)

        self.changed_callbacks = CallbackList()

    def getdict(self, names = None):
        raise NotImplementedError()

    def get(self, name, default = None):
        raise NotImplementedError()

    def set(self, name, value, reply_handler, error_handler = None):
        raise NotImplementedError()

    ###

    def wait_for_value(self, name, value, timeout = DEFAULT_TIMEOUT):
        # FIXME: Check that we are not running in the main thread.

        queue = Queue.Queue()

        f = lambda *a: queue.put(a)

        self.changed_callbacks.add(f)
        try:
            props = self.getdict()
            if props.get(name, None) == value:
                return props
            while True:
                try:
                    parent, names, props = queue.get(timeout = timeout)
                except Queue.Empty:
                    raise TimeoutError()
                if props.get(name, None) == value:
                    return props
        finally:
            self.changed_callbacks.remove(f)


class DBusPropertiesObject(PropertiesObject, DBusObject):
    properties_iface = None

    _properties = None
    _lock = None

    def __init__(self, properties_iface, **kwargs):
        self.properties_iface = properties_iface

        self._lock = RLock()

        super(DBusPropertiesObject, self).__init__(**kwargs)

        self.connect_to_signal(
          'PropertiesChanged',
          self._handle_properties_changed,
        )
        self.call(
          'GetAll',
          self.properties_iface,
          reply_handler = self._handle_get_all_reply_init,
        )

    def get_proxy_object(self):
        obj = super(DBusPropertiesObject, self).get_proxy_object()
        return dbus.Interface(obj, 'org.freedesktop.DBus.Properties')

    def is_ready(self):
        if not super(DBusPropertiesObject, self).is_ready():
            return False
        return (self._properties is not None)

    def _handle_get_all_reply_init(self, properties):
        self._properties = properties
        self.maybe_ready()

    def _handle_properties_changed(
      self,
      interface_name,
      changed_properties,
      invalidated_properties,
    ):
        if invalidated_properties:
            raise AssertionError(
              "don't know how to handle invalidated_properties"
            )

        self._maybe_properties_changed(changed_properties)

    # FIXME: Confirm that the following is true.

    # We really shouldn't need this, but there is some odd behavior out there
    # that requires it (udisks).

    def refresh(self, callback = None):
        self.call(
          'GetAll',
          self.properties_iface,
          reply_handler = self._handle_get_all_reply_refresh,
        )

    def _handle_get_all_reply_refresh(self, properties):
        self._maybe_properties_changed(properties)

    def _maybe_properties_changed(self, changed_properties):
        _changed_properties = dict(changed_properties)
        if self._properties is not None:
            with self._lock:
                for name, value in changed_properties.items():
                    old_value = self._properties[name]
                    if value != old_value:
                        self._properties[name] = value
                    else:
                        del _changed_properties[name]
                property_names = _changed_properties.keys()
                properties = dict(self._properties)

            if property_names:
                self.changed_callbacks.call(
                  self.parent,
                  property_names,
                  properties,
                )

    ###

    def getdict(self, names = None):
        with self._lock:
            if names is None:
                return dict(self._properties)
            else:
                return dict((name, self._properties[name]) for name in names)

    def get(self, name, default = None):
        with self._lock:
            if default is not None:
                return self._properties.get(name, default)
            return self._properties[name]

    def set(self, name, value, reply_handler, error_handler = None):
        kwargs = {'reply_handler': reply_handler}
        if error_handler is not None:
            kwargs['error_handler'] = error_handler
        self.call('Set', self.properties_iface, name, value, **kwargs)


class DBusObjectWithDBusProperties(DBusObject):
    properties = None

    def __init__(self, properties_iface, **kwargs):
        super(DBusObjectWithDBusProperties, self).__init__(**kwargs)

        self.properties = DBusPropertiesObject(
          bus = self.bus,
          parent = self,
          bus_name = self.bus_name,
          object_path = self.object_path,
          properties_iface = properties_iface,
          ready_callback = self._properties_ready,
          default_error_handler = self.default_error_handler,
        )

    def is_ready(self):
        if not super(DBusObjectWithDBusProperties, self).is_ready():
            return False
        return (self.properties is not None) and self.properties.is_ready()

    def _properties_ready(self, properties):
        self.maybe_ready()


###


class DBusObjectSet(Object):
    object_cls = None

    object_added_callbacks = None
    object_removed_callbacks = None

    _lock = None
    _object_paths = None
    _objects = None
    _is_ready = None

    def __init__(self, object_cls, **kwargs):
        self.object_cls = object_cls

        self._lock = RLock()
        self._objects = {}
        self._is_ready = False

        super(DBusObjectSet, self).__init__(**kwargs)

        self.object_added_callbacks = CallbackList()
        self.object_removed_callbacks = CallbackList()

    def is_ready(self):
        # Our ready state can fluctuate after initially being true, but we
        # consider ourselves ready if we have been ready at some time in
        # the past.
        if self._is_ready:
            return True
        if not super(DBusObjectSet, self).is_ready():
            return False
        with self._lock:
            is_ready = (
              (self._object_paths is not None)
              and (set(self._objects.keys()) == self._object_paths)
            )
            if is_ready:
                self._is_ready = True
            return is_ready

    def add_initial_objects(self, object_paths):
        with self._lock:
            self._object_paths = set([])
            for object_path in object_paths:
                self.add_object(object_path)

        self.maybe_ready()

    def add_object(self, object_path):
        if self._object_paths is None:
            return

        with self._lock:
            self._object_paths.add(object_path)

        self.object_cls(
          bus = self.bus,
          object_path = object_path,
          ready_callback = self._handle_object_ready,
          default_error_handler = self.default_error_handler,
        )

        self.maybe_ready()

    def _handle_object_ready(self, obj):
        was_ready = self.is_ready()

        added = False
        with self._lock:
            if (
              obj.object_path in self._object_paths
              and obj.object_path not in self._objects
            ):
                self._objects[obj.object_path] = obj
                added = True

        if added and was_ready:
            self.object_added_callbacks.call(obj)

        self.maybe_ready()

    def remove_object(self, object_path):
        if self._object_paths is None:
            return

        removed = False
        with self._lock:
            self._object_paths.discard(object_path)
            if object_path in self._objects:
                obj = self._objects.pop(object_path)
                removed = True

        if removed:
            self.object_removed_callbacks.call(obj)
            obj.disconnect_signals()

        self.maybe_ready()

    def iter_objects(self):
        with self._lock:
            for obj in self._objects.itervalues():
                yield obj

    def get_object(self, object_path):
        with self._lock:
            return self._objects[object_path]


class DBusObjectList(Object):
    object_cls = None

    object_added_callbacks = None
    object_removed_callbacks = None

    _lock = None
    _object_paths = None
    _objects = None
    _is_ready = None

    def __init__(self, object_cls, **kwargs):
        self.object_cls = object_cls

        self._lock = RLock()
        self._objects = {}
        self._is_ready = False

        super(DBusObjectList, self).__init__(**kwargs)

        self.object_added_callbacks = CallbackList()
        self.object_removed_callbacks = CallbackList()

    def is_ready(self):
        # Our ready state can fluctuate after initially being true, but we
        # consider ourselves ready if we have been ready at some time in
        # the past.
        if self._is_ready:
            return True
        if not super(DBusObjectList, self).is_ready():
            return False
        with self._lock:
            is_ready = (
              (self._object_paths is not None)
              and (set(self._objects.keys()) == set(self._object_paths))
            )
            if is_ready:
                self._is_ready = True
            return is_ready

    def add_initial_objects(self, object_paths):
        with self._lock:
            self._object_paths = list(object_paths)
            for object_path in object_paths:
                self._fetch_object(object_path)
        self.maybe_ready()

    def add_object(self, index, object_path):
        if self._object_paths is None:
            return

        with self._lock:
            self._object_paths.insert(index, object_path)

        self._fetch_object(object_path)
        self.maybe_ready()

    def _fetch_object(self, object_path):
        self.object_cls(
          bus = self.bus,
          object_path = object_path,
          ready_callback = self._handle_object_ready,
          default_error_handler = self.default_error_handler,
        )

    def _handle_object_ready(self, obj):
        was_ready = self.is_ready()

        added = False
        with self._lock:
            if (
              obj.object_path in self._object_paths
              and obj.object_path not in self._objects
            ):
                self._objects[obj.object_path] = obj
                added = True

        if added and was_ready:
            index = self._object_paths.index(obj.object_path)
            self.object_added_callbacks.call(index, obj)

        self.maybe_ready()

    def remove_object(self, object_path):
        if self._object_paths is None:
            return

        removed = False
        with self._lock:
            try:
                index = self._object_paths.index(object_path)
            except ValueError:
                pass
            else:
                del self._object_paths[index]
                if object_path in self._objects:
                    obj = self._objects.pop(object_path)
                    removed = True

        if removed:
            self.object_removed_callbacks.call(index, obj)
            obj.disconnect_signals()

        self.maybe_ready()

    def iter_objects(self):
        with self._lock:
            for object_path in self._object_paths:
                try:
                    yield self._objects[object_path]
                except KeyError:
                    pass

    def get_object(self, object_path):
        with self._lock:
            return self._objects[object_path]

    def update_objects(self, object_paths):
        with self._lock:
            added = []

            for object_path in self._object_paths:
                if object_path not in object_paths:
                    self.remove_object(object_path)

            for object_path in object_paths:
                if object_path not in self._object_paths:
                    added.append(object_path)

            self._object_paths[:] = object_paths

            for object_path in added:
                self._fetch_object(object_path)
